# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import falcon

from scout_apm.api import install
from scout_apm.core.ignore import ignore_path
from scout_apm.core.queue_time import track_request_queue_time
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)

# Falcon Middleware docs:
# https://falcon.readthedocs.io/en/stable/api/middleware.html


class ScoutMiddleware(object):
    """
    Falcon Middleware for integration with Scout APM.
    """

    def __init__(self, config):
        install(config=config)
        self.api = None

    def set_api(self, api):
        if not isinstance(api, falcon.API):
            raise ValueError("api should be an instance of falcon.API")
        self.api = api

    def process_request(self, req, resp):
        tracked_request = TrackedRequest.instance()
        tracked_request.mark_real_request()
        req.context.scout_tracked_request = tracked_request

        tracked_request.tag("path", req.path)
        if ignore_path(req.path):
            tracked_request.tag("ignore_transaction", True)

        # Determine a remote IP to associate with the request. The value is
        # spoofable by the requester so this is not suitable to use in any
        # security sensitive context.
        user_ip = (
            req.get_header("x-forwarded-for", default="").split(",")[0]
            or req.get_header("client-ip", default="").split(",")[0]
            or req.remote_addr
        )
        tracked_request.tag("user_ip", user_ip)

        queue_time = req.get_header("x-queue-start", default="") or req.get_header(
            "x-request-start", default=""
        )
        track_request_queue_time(queue_time, tracked_request)

    def process_resource(self, req, resp, resource, params):
        tracked_request = getattr(req.context, "scout_tracked_request", None)
        if tracked_request is None:
            # Somehow we didn't start a request - this might occur in
            # combination with a pretty adversarial application, so guard
            # against it, although if a request was started and the context was
            # lost, other problems might occur.
            return

        if self.api is None:
            logger.warning(
                (
                    "{}.set_api() should be called before requests begin for "
                    "more detail"
                ).format(self.__class__.__name__)
            )
            operation = "Controller/{}.{}.{}".format(
                resource.__module__, resource.__class__.__name__, req.method
            )
        else:
            # Find the current responder's name. Falcon passes middleware the
            # current resource but unfortunately not the method being called, hence
            # we have to go through routing again.
            responder, _params, _resource, _uri_template = self.api._get_responder(req)
            operation = "Controller/{}.{}.{}".format(
                resource.__module__, resource.__class__.__name__, responder.__name__
            )

        span = tracked_request.start_span(operation=operation)
        req.context.scout_resource_span = span

    def process_response(self, req, resp, resource, req_succeeded):
        tracked_request = getattr(req.context, "scout_tracked_request", None)
        if tracked_request is None:
            # Somehow we didn't start a request
            return

        if not req_succeeded:
            tracked_request.tag("error", "true")

        span = getattr(req.context, "scout_resource_span", None)
        if span is not None:
            tracked_request.stop_span()
        else:
            tracked_request.finish()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import psutil

logger = logging.getLogger(__name__)


class Memory(object):
    @staticmethod
    def rss_to_mb(rss_in_bytes):
        """
        Convert a number of bytes to a number of megabytes.
        """
        bytes_per_kb = 1024
        kb_per_megabyte = 1024
        return float(rss_in_bytes) / bytes_per_kb / kb_per_megabyte

    @staticmethod
    def rss():
        """
        Returns the memory usage (RSS) of this process, in bytes
        """
        return psutil.Process().memory_info().rss

    @staticmethod
    def rss_in_mb():
        return Memory.rss_to_mb(Memory.rss())

    @staticmethod
    def get_delta(prior_rss_in_mb):
        mem = Memory.rss_in_mb()
        if mem > prior_rss_in_mb:
            return mem - prior_rss_in_mb
        return 0.0

    def metric_type(self):
        return "Memory"

    def metric_name(self):
        return "Physical"

    def human_name(self):
        return "Process Memory"

    def run(self):
        res = self.__class__.rss_in_mb()
        logger.debug("%s: #%s", self.human_name(), res)
        return res



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

from flask import current_app
from flask.globals import _request_ctx_stack

import scout_apm.core
from scout_apm.core.config import ScoutConfig
from scout_apm.core.ignore import ignore_path
from scout_apm.core.monkey import CallableProxy
from scout_apm.core.queue_time import track_request_queue_time
from scout_apm.core.tracked_request import TrackedRequest


class ScoutApm(object):
    def __init__(self, app):
        self.app = app
        app.before_first_request(self.before_first_request)
        # Wrap the Flask.dispatch_request method
        app.dispatch_request = CallableProxy(
            app.dispatch_request, self.wrapped_dispatch_request
        )

    #############
    #  Startup  #
    #############

    def before_first_request(self):
        self.extract_flask_settings()
        scout_apm.core.install()

    def extract_flask_settings(self):
        """
        Copies SCOUT_* settings in the app into Scout's config lookup
        """
        configs = {}
        configs["application_root"] = self.app.instance_path
        for name in current_app.config:
            if name.startswith("SCOUT_"):
                value = current_app.config[name]
                clean_name = name.replace("SCOUT_", "").lower()
                configs[clean_name] = value
        ScoutConfig.set(**configs)

    ############################
    #  Request Lifecycle hook  #
    ############################

    def wrapped_dispatch_request(self, original, *args, **kwargs):
        request = _request_ctx_stack.top.request

        # Copied logic from Flask
        if request.routing_exception is not None:
            return original(*args, **kwargs)

        rule = request.url_rule
        view_func = self.app.view_functions[rule.endpoint]

        path = request.path
        name = view_func.__module__ + "." + view_func.__name__
        operation = "Controller/" + name

        tracked_request = TrackedRequest.instance()
        tracked_request.mark_real_request()
        tracked_request.tag("path", path)
        span = tracked_request.start_span(operation=operation)
        span.tag("name", name)

        if ignore_path(path):
            tracked_request.tag("ignore_transaction", True)

        # Determine a remote IP to associate with the request. The value is
        # spoofable by the requester so this is not suitable to use in any
        # security sensitive context.
        user_ip = (
            request.headers.get("x-forwarded-for", default="").split(",")[0]
            or request.headers.get("client-ip", default="").split(",")[0]
            or request.remote_addr
        )
        tracked_request.tag("user_ip", user_ip)

        queue_time = request.headers.get(
            "x-queue-start", default=""
        ) or request.headers.get("x-request-start", default="")
        track_request_queue_time(queue_time, tracked_request)

        try:
            return original(*args, **kwargs)
        except Exception as exc:
            tracked_request.tag("error", "true")
            raise exc
        finally:
            tracked_request.stop_span()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from scout_apm.core.monkey import monkeypatch_method, unpatch_method
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)


class Instrument(object):
    installed = False

    def installable(self):
        try:
            from jinja2 import Template  # noqa: F401
        except ImportError:
            logger.info("Unable to import for Jinja2 instruments")
            return False
        if self.installed:
            logger.warning("Jinja2 Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("Jinja2 instruments are not installable. Skipping.")
            return False

        self.__class__.installed = True

        try:
            from jinja2 import Template

            @monkeypatch_method(Template)
            def render(original, self, *args, **kwargs):
                tr = TrackedRequest.instance()
                span = tr.start_span(operation="Template/Render")
                span.tag("name", self.name)

                try:
                    return original(*args, **kwargs)
                finally:
                    tr.stop_span()

            logger.info("Instrumented Jinja2")

        except Exception as e:
            logger.warning("Unable to instrument for Jinja2 Template.render: %r", e)
            return False
        return True

    def uninstall(self):
        if not self.installed:
            logger.info("Jinja2 instruments are not installed. Skipping.")
            return False

        self.__class__.installed = False

        from jinja2 import Template

        unpatch_method(Template, "render")



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import scout_apm.core
from scout_apm.compat import ContextDecorator, text
from scout_apm.core.config import ScoutConfig
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)


__all__ = [
    "BackgroundTransaction",
    "Config",
    "Context",
    "WebTransaction",
    "install",
    "instrument",
]


class Context(object):
    @classmethod
    def add(self, key, value):
        """Adds context to the currently executing request.

        :key: Any String identifying the request context.
              Example: "user_ip", "plan", "alert_count"
        :value: Any json-serializable type.
              Example: "1.1.1.1", "free", 100
        :returns: nothing.
        """
        TrackedRequest.instance().tag(key, value)


class Config(ScoutConfig):
    pass


def install(*args, **kwargs):
    scout_apm.core.install(*args, **kwargs)


def ignore_transaction():
    TrackedRequest.instance().tag("ignore_transaction", True)


class instrument(ContextDecorator):
    def __init__(self, operation, kind="Custom", tags={}):
        self.operation = text(kind) + "/" + text(operation)
        self.tags = tags

    def __enter__(self):
        tr = TrackedRequest.instance()
        self.span = tr.start_span(operation=self.operation)
        for key, value in self.tags.items():
            self.tag(key, value)
        return self

    def __exit__(self, *exc):
        tr = TrackedRequest.instance()
        tr.stop_span()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)


class Transaction(ContextDecorator):
    """
    This Class is not meant to be used directly.
    Use one of the subclasses
    (WebTransaction or BackgroundTransaction)
    """

    def __init__(self, name, tags={}):
        self.name = text(name)
        self.tags = tags

    @classmethod
    def start(cls, kind, name, tags={}):
        operation = text(kind) + "/" + text(name)

        tr = TrackedRequest.instance()
        tr.mark_real_request()
        span = tr.start_span(operation=operation)
        for key, value in tags.items():
            tr.tag(key, value)
        return span

    @classmethod
    def stop(cls):
        tr = TrackedRequest.instance()
        tr.stop_span()
        return True

    # __enter__ must be defined by child classes.

    # *exc is any exception raised. Ignore that
    def __exit__(self, *exc):
        WebTransaction.stop()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)


class WebTransaction(Transaction):
    @classmethod
    def start(cls, name, tags={}):
        Transaction.start("Controller", text(name), tags)

    def __enter__(self):
        Transaction.start("Controller", self.name, self.tags)


class BackgroundTransaction(Transaction):
    @classmethod
    def start(cls, name, tags={}):
        Transaction.start("Job", text(name), tags)

    def __enter__(self):
        Transaction.start("Job", self.name, self.tags)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

logger = logging.getLogger(__name__)


class NPlusOneCallSet(object):
    def __init__(self):
        self.items = {}

    def update(self, sql_string, call_count, call_duration):
        self.find_for(sql_string).update(call_count, call_duration)

    def should_capture_backtrace(self, sql_string):
        return self.find_for(sql_string).should_capture_backtrace()

    def find_for(self, sql_string):
        if sql_string not in self.items:
            self.items[sql_string] = NPlusOneCallSetItem(sql_string)
        return self.items[sql_string]


class NPlusOneCallSetItem(object):
    # Fetch backtraces on this number of same SQL calls
    CALL_COUNT_THRESHOLD = 5

    # Minimum time in seconds before we start performing any work.
    DURATION_THRESHOLD = 150 / 1000.0

    def __init__(self, sql_string):
        self.sql_string = sql_string
        self.captured = False
        self.call_count = 0
        self.call_duration = 0.0  # In Seconds

    def update(self, call_count, call_duration):
        if self.captured:
            # No need to do any work if we've already captured a backtrace.
            return
        self.call_count += call_count
        self.call_duration += call_duration

    def should_capture_backtrace(self):
        if self.captured:
            return False
        # Call count and call duration must both be past thresholds.
        if (
            self.call_count >= self.CALL_COUNT_THRESHOLD
            and self.call_duration >= self.DURATION_THRESHOLD
        ):
            self.captured = True
            return True
        return False



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from scout_apm.core.monkey import monkeypatch_method, unpatch_method
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)


class Instrument(object):
    installed = False

    def installable(self):
        try:
            from urllib3 import HTTPConnectionPool, PoolManager  # noqa: F401
        except ImportError:
            logger.info("Unable to import for Urllib3 instruments")
            return False
        if self.installed:
            logger.warning("Urllib3 Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("Urllib3 instruments are not installable. Skipping.")
            return False

        self.__class__.installed = True

        try:
            from urllib3 import HTTPConnectionPool

            @monkeypatch_method(HTTPConnectionPool)
            def urlopen(original, self, *args, **kwargs):
                method = "Unknown"
                url = "Unknown"
                try:
                    try:
                        method = kwargs["method"]
                    except KeyError:
                        method = args[0]
                    url = "{}".format(self._absolute_url("/"))
                except Exception:
                    logger.exception(
                        "Could not get instrument data for HTTPConnectionPool"
                    )

                tr = TrackedRequest.instance()
                span = tr.start_span(operation="HTTP/{}".format(method))
                span.tag("url", "{}".format(url))

                try:
                    return original(*args, **kwargs)
                finally:
                    tr.stop_span()

            logger.info("Instrumented Urllib3")

        except Exception as e:
            logger.warning(
                "Unable to instrument for Urllib3 HTTPConnectionPool.urlopen: %r", e
            )
            return False
        return True

    def uninstall(self):
        if not self.installed:
            logger.info("Urllib3 instruments are not installed. Skipping.")
            return False

        self.__class__.installed = False

        from urllib3 import HTTPConnectionPool

        unpatch_method(HTTPConnectionPool, "urlopen")



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime as dt
import sys

string_type = str if sys.version_info[0] >= 3 else basestring  # noqa: F821
text_type = str if sys.version_info[0] >= 3 else unicode  # noqa: F821

# Python 2 (and very early 3.x) didn't have ContextDecorator, so define it for ourselves
if sys.version_info < (3, 2):
    import functools

    class ContextDecorator(object):
        def __call__(self, f):
            @functools.wraps(f)
            def decorated(*args, **kwds):
                with self:
                    return f(*args, **kwds)

            return decorated


else:
    from contextlib import ContextDecorator

try:
    # Python 3.x
    import queue
except ImportError:
    # Python 2.x
    import Queue as queue

# datetime_to_timestamp converts a naive UTC datetime to a unix timestamp
if sys.version_info >= (3, 3):

    def datetime_to_timestamp(datetime_obj):
        return datetime_obj.replace(tzinfo=dt.timezone.utc).timestamp()


else:
    _EPOCH = dt.datetime(1970, 1, 1)

    def datetime_to_timestamp(datetime_obj):
        return (datetime_obj - _EPOCH).total_seconds()


def text(value, encoding="utf-8", errors="strict"):
    """
    Convert a value to str on Python 3 and unicode on Python 2.
    """
    if isinstance(value, text_type):
        return value
    elif isinstance(value, bytes):
        return text_type(value, encoding, errors)
    else:
        return text_type(value)


__all__ = [
    "ContextDecorator",
    "datetime_to_timestamp",
    "queue",
    "string_type",
    "text",
    "text_type",
]



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime as dt

from celery.signals import before_task_publish, task_postrun, task_prerun

import scout_apm.core
from scout_apm.compat import datetime_to_timestamp
from scout_apm.core.tracked_request import TrackedRequest


def before_publish_callback(headers=None, properties=None, **kwargs):
    if "scout_task_start" not in headers:
        headers["scout_task_start"] = datetime_to_timestamp(dt.datetime.utcnow())


def prerun_callback(task=None, **kwargs):
    tracked_request = TrackedRequest.instance()
    tracked_request.mark_real_request()

    start = getattr(task.request, "scout_task_start", None)
    if start is not None:
        now = datetime_to_timestamp(dt.datetime.utcnow())
        try:
            queue_time = now - start
        except TypeError:
            pass
        else:
            tracked_request.tag("queue_time", queue_time)

    task_id = getattr(task.request, "id", None)
    if task_id:
        tracked_request.tag("task_id", task_id)
    parent_task_id = getattr(task.request, "parent_id", None)
    if parent_task_id:
        tracked_request.tag("parent_task_id", parent_task_id)

    delivery_info = task.request.delivery_info
    tracked_request.tag("is_eager", delivery_info.get("is_eager", False))
    tracked_request.tag("exchange", delivery_info.get("exchange", "unknown"))
    tracked_request.tag("routing_key", delivery_info.get("routing_key", "unknown"))
    tracked_request.tag("queue", delivery_info.get("queue", "unknown"))

    tracked_request.start_span(operation=("Job/" + task.name))


def postrun_callback(task=None, **kwargs):
    tracked_request = TrackedRequest.instance()
    tracked_request.stop_span()


def install():
    installed = scout_apm.core.install()
    if not installed:
        return

    before_task_publish.connect(before_publish_callback)
    task_prerun.connect(prerun_callback)
    task_postrun.connect(postrun_callback)


def uninstall():
    before_task_publish.disconnect(before_publish_callback)
    task_prerun.disconnect(prerun_callback)
    task_postrun.disconnect(postrun_callback)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

# XXX:Changed in Django 1.9
# https://github.com/jazzband/django-debug-toolbar/issues/739
#  from django.template.base import Library, Template
from django.template import Library, Template, defaulttags
from django.template.loader_tags import BlockNode

from scout_apm.core.stacktracer import trace_function, trace_method

logger = logging.getLogger(__name__)

register = Library()


class DecoratingParserProxy(object):  # pragma: no cover
    """
    Mocks out the django template parser, passing templatetags through but
    first wrapping them to include performance data
    """

    def __init__(self, parser):
        self.parser = parser

    def add_library(self, library):
        wrapped_library = Library()
        wrapped_library.filters = library.filters
        for name, tag_compiler in library.tags.items():
            wrapped_library.tags[name] = self.wrap_compile_function(name, tag_compiler)
        self.parser.add_library(wrapped_library)

    def wrap_compile_function(self, name, tag_compiler):
        def compile(*args, **kwargs):
            node = tag_compiler(*args, **kwargs)
            node.render = trace_function(node.render, ("Template/Tag", {"name": name}))
            return node

        return compile


@register.tag
def load(parser, token):  # pragma: no cover
    decorating_parser = DecoratingParserProxy(parser)
    return defaulttags.load(decorating_parser, token)


def install_template_instrumentation():
    if getattr(install_template_instrumentation, "installed", False):
        return
    install_template_instrumentation.installed = True

    # Our eventual aim is to patch the render() method on the Node objects
    # corresponding to custom template tags. However we have to jump through
    # a number of hoops in order to get access to the object.
    #
    # 1. Add ourselves to the set of built in template tags
    #    This allows us to replace the 'load' template tag which controls
    #    the loading of custom template tags
    # 2. Delegate to default load with replaced parser
    #    We provide our own parser class so we can catch and intercept
    #    calls to add_library.
    # 3. add_library receives a library of template tags
    #    It iterates through each template tag, wrapping its compile function
    # 4. compile is called as part of compiling the template
    #    Our wrapper is called instead of the original templatetag compile
    #    function. We delegate to the original function, but then modify
    #    the resulting object by wrapping its render() function. This
    #    render() function is what ends up being timed and appearing in the
    #    tree.

    # XXX: Stopped working in Django 1.9
    #  add_to_builtins('apm.instruments.view')

    @trace_method(Template)
    def __init__(self, *args, **kwargs):
        name = args[2] if len(args) >= 3 else "<Unknown Template>"
        return ("Template/Compile", {"name": name})

    @trace_method(Template)
    def render(self, *args, **kwargs):
        name = self.name if self.name is not None else "<Unknown Template>"
        return ("Template/Render", {"name": name})

    @trace_method(BlockNode)  # noqa: F811
    def render(self, *args, **kwargs):
        return ("Block/Render", {"name": self.name})

    logger.debug("Monkey patched Templates")

    # XXX: Figure this out, causes exception that the "resolve_context" key
    # isn't in dict
    # Also will need to figure out the name hash
    #  @trace_method(TemplateResponse)
    #  def resolve_context(self, *args, **kwargs):
    #  return ('Template/Context', {})



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals


def octal(value):
    """
    Takes an integer or a string containing an integer and returns the octal
    value. Raises a ValueError if the value cannot be converted to octal.
    """
    return int("{}".format(value), 8)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import time

from scout_apm.compat import datetime_to_timestamp


def track_request_queue_time(header_value, tracked_request):
    if header_value.startswith("t="):
        header_value = header_value[2:]

    try:
        first_char = header_value[0]
    except IndexError:
        return

    if not first_char.isdigit():  # filter out negatives, nan, inf, etc.
        return

    try:
        ambiguous_start_timestamp = float(header_value)
    except ValueError:
        return

    start_timestamp_ns = convert_ambiguous_timestamp_to_ns(ambiguous_start_timestamp)
    if start_timestamp_ns == 0.0:
        return

    tr_start_timestamp_ns = datetime_to_timestamp(tracked_request.start_time) * 1e9

    # Ignore if in the future
    if start_timestamp_ns > tr_start_timestamp_ns:
        return

    queue_time_ns = int(tr_start_timestamp_ns - start_timestamp_ns)
    tracked_request.tag("scout.queue_time_ns", queue_time_ns)


# Cutoff epoch is used for determining ambiguous timestamp boundaries, and is
# just over 10 years ago at time of writing
CUTOFF_EPOCH_S = time.mktime((2009, 6, 1, 0, 0, 0, 0, 0, 0))
CUTOFF_EPOCH_MS = CUTOFF_EPOCH_S * 1000.0
CUTOFF_EPOCH_US = CUTOFF_EPOCH_S * 1000000.0
CUTOFF_EPOCH_NS = CUTOFF_EPOCH_S * 1000000000.0


def convert_ambiguous_timestamp_to_ns(timestamp):
    """
    Convert an ambiguous float timestamp that could be in nanoseconds,
    microseconds, milliseconds, or seconds to nanoseconds. Return 0.0 for
    values in the more than 10 years ago.
    """
    if timestamp > CUTOFF_EPOCH_NS:
        converted_timestamp = timestamp
    elif timestamp > CUTOFF_EPOCH_US:
        converted_timestamp = timestamp * 1000.0
    elif timestamp > CUTOFF_EPOCH_MS:
        converted_timestamp = timestamp * 1000000.0
    elif timestamp > CUTOFF_EPOCH_S:
        converted_timestamp = timestamp * 1000000000.0
    else:
        return 0.0
    return converted_timestamp



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging
import os
import socket
import struct
import threading
import time

from scout_apm.compat import queue
from scout_apm.core.commands import Register
from scout_apm.core.config import ScoutConfig

SECOND = 1  # time unit - monkey-patched in tests to make them run faster

logger = logging.getLogger(__name__)


class CoreAgentSocket(threading.Thread):
    _instance = None
    _instance_lock = threading.Lock()

    @classmethod
    def instance(cls, *args, **kwargs):
        with cls._instance_lock:
            # No instance exists yet.
            if cls._instance is None:
                cls._instance = cls(args, kwargs)
                return cls._instance

            # An instance exists but is no longer running.
            if not cls._instance.running():
                cls._instance = cls(args, kwargs)
                return cls._instance

            # An instance exists and is running (or in the process of
            # starting or in the process of stopping). In any case,
            # return this instance.
            return cls._instance

    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self)
        self.config = kwargs.get("scout_config", ScoutConfig())

        # Socket related
        self.socket_path = self.config.value("socket_path")
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        # Threading control related
        self._started_event = threading.Event()
        self._stop_event = threading.Event()
        self._stopped_event = threading.Event()

        # Command queues
        self.command_queue = queue.Queue(maxsize=500)

        # Set Thread options
        self.daemon = True

        # Set the started event here to avoid races in the class instance()
        # method. If there is an exception in the socket thread then it will
        # clear this event on exit.
        self._started_event.set()

        # Now call start() which eventually launches run() in another thread.
        self.start()

    def __del__(self):
        self.stop()

    def running(self):
        return self._started_event.is_set()

    def stop(self):
        if self._started_event.is_set():
            self._stop_event.set()
            self.command_queue.put(None, False)  # unblock self.command_queue.get
            self._stopped_event.wait(2 * SECOND)
            if self._stopped_event.is_set():
                return True
            else:
                logger.debug("CoreAgentSocket Failed to stop thread within timeout!")
                return False
        else:
            return True

    def run(self):
        """
        Called by the threading system
        """

        try:
            self._connect()
            self._register()
            while True:
                try:
                    body = self.command_queue.get(block=True, timeout=1 * SECOND)
                except queue.Empty:
                    body = None

                if body is not None:
                    result = self._send(body)
                    if result:
                        self.command_queue.task_done()
                    else:
                        # Something was wrong with the socket.
                        self._disconnect()
                        self._connect()
                        self._register()

                # Check for stop event after a read from the queue. This is to
                # allow you to open a socket, immediately send to it, and then
                # stop it. We do this in the Metadata send at application start
                # time
                if self._stop_event.is_set():
                    logger.debug("CoreAgentSocket thread stopping.")
                    break
        except Exception:
            logger.debug("CoreAgentSocket thread exception.")
        finally:
            self._started_event.clear()
            self._stop_event.clear()
            self._stopped_event.set()
            logger.debug("CoreAgentSocket thread stopped.")

    def send(self, command):
        try:
            self.command_queue.put(command, False)
        except queue.Full as e:
            # TODO mark the command as not queued?
            logger.debug("CoreAgentSocket error on send: %r", e)

    def _send(self, command):
        msg = command.message()

        try:
            data = json.dumps(msg)
        except (ValueError, TypeError) as e:
            logger.debug("Exception when serializing command message: %r", e)
            return False

        try:
            self.socket.sendall(self._message_length(data))
        except OSError as e:
            logger.debug(
                "CoreAgentSocket exception on length _send: "
                "%r on PID: %s on thread: %s",
                e,
                os.getpid(),
                threading.current_thread(),
            )
            return None

        try:
            self.socket.sendall(data.encode())
        except OSError as e:
            logger.debug(
                "CoreAgentSocket exception on data _send: "
                "%r on PID: %s on thread: %s",
                e,
                os.getpid(),
                threading.current_thread(),
            )
            return None

        # TODO do something with the response sent back in reply to command
        self._read_response()

        return True

    def _message_length(self, body):
        length = len(body)
        return struct.pack(">I", length)

    def _read_response(self):
        try:
            raw_size = self.socket.recv(4)
            size = struct.unpack(">I", raw_size)[0]
            message = bytearray(0)

            while len(message) < size:
                recv = self.socket.recv(size)
                message += recv

            return message
        except OSError as e:
            logger.debug("CoreAgentSocket error on read response: %r", e)
            return None

    def _register(self):
        self._send(
            Register(app=self.config.value("name"), key=self.config.value("key"))
        )

    def _connect(self, connect_attempts=5, retry_wait_secs=1):
        for attempt in range(1, connect_attempts + 1):
            logger.debug(
                "CoreAgentSocket attempt %d, connecting to %s, PID: %s, Thread: %s",
                attempt,
                self.socket_path,
                os.getpid(),
                threading.current_thread(),
            )
            try:
                self.socket.connect(self.socket_path)
                self.socket.settimeout(0.5 * SECOND)
                logger.debug("CoreAgentSocket is connected")
                return True
            except socket.error as e:
                logger.debug("CoreAgentSocket connection error: %r", e)
                # Return without waiting when reaching the maximum number of attempts.
                if attempt >= connect_attempts:
                    return False
                time.sleep(retry_wait_secs * SECOND)

    def _disconnect(self):
        logger.debug("CoreAgentSocket disconnecting from %s", self.socket_path)
        try:
            self.socket.close()
        except socket.error as e:
            logger.debug("CoreAgentSocket exception on disconnect: %r", e)
        finally:
            self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import django
from django.db import connections
from django.db.backends.signals import connection_created

from scout_apm.core.monkey import monkeypatch_method
from scout_apm.core.tracked_request import TrackedRequest

try:
    from django.db.backends.base.base import BaseDatabaseWrapper
except ImportError:
    # Backwards compatibility for Django <1.8
    from django.db.backends import BaseDatabaseWrapper

try:
    from django.db.backends.utils import CursorWrapper
except ImportError:
    # Backwards compatibility for Django <1.9
    from django.db.backends.util import CursorWrapper

logger = logging.getLogger(__name__)


def db_execute_hook(execute, sql, params, many, context):
    """
    Database instrumentation hook for Django 2.0+
    https://docs.djangoproject.com/en/2.0/topics/db/instrumentation/
    """
    if many:
        operation = "SQL/Many"
    else:
        operation = "SQL/Query"

    tracked_request = TrackedRequest.instance()
    span = tracked_request.start_span(operation=operation)
    span.tag("db.statement", sql)

    try:
        return execute(sql, params, many, context)
    finally:
        tracked_request.stop_span()
        tracked_request.callset.update(sql, 1, span.duration())
        if tracked_request.callset.should_capture_backtrace(sql):
            span.capture_backtrace()


def install_db_execute_hook(connection, **kwargs):
    """
    Install db_execute_hook on the given database connection.

    Rather than use the documented API of the `execute_wrapper()` context
    manager, directly insert the hook. This is done because:
    1. Deleting the context manager closes it, so it's not possible to enter
       it here and not exit it, unless we store it forever in some variable
    2. We want to be the first hook, so we can capture every query (although
       potentially later hooks will change the SQL)
    3. We want to be idempotent and only install the hook once
    """
    if db_execute_hook not in connection.execute_wrappers:
        connection.execute_wrappers.insert(0, db_execute_hook)


class _DetailedTracingCursorWrapper(CursorWrapper):
    """
    Monkey-patched-in cursor wrapper for Django < 2.0
    """

    def execute(self, sql, params=None):
        tracked_request = TrackedRequest.instance()
        span = tracked_request.start_span(operation="SQL/Query")
        span.tag("db.statement", sql)

        try:
            return self.cursor.execute(sql, params)
        finally:
            tracked_request.stop_span()
            tracked_request.callset.update(sql, 1, span.duration())
            if tracked_request.callset.should_capture_backtrace(sql):
                span.capture_backtrace()

    def executemany(self, sql, param_list):
        tracked_request = TrackedRequest.instance()
        span = tracked_request.start_span(operation="SQL/Many")
        span.tag("db.statement", sql)

        try:
            return self.cursor.executemany(sql, param_list)
        finally:
            tracked_request.stop_span()
            tracked_request.callset.update(sql, 1, span.duration())
            if tracked_request.callset.should_capture_backtrace(sql):
                span.capture_backtrace()


def install_sql_instrumentation():
    if getattr(install_sql_instrumentation, "installed", False):
        return
    install_sql_instrumentation.installed = True

    if django.VERSION >= (2, 0):
        for connection in connections.all():
            install_db_execute_hook(connection=connection)
        connection_created.connect(install_db_execute_hook)
        logger.debug("Installed DB connection created signal handler")
    else:

        @monkeypatch_method(BaseDatabaseWrapper)
        def cursor(original, self, *args, **kwargs):
            result = original(*args, **kwargs)
            return _DetailedTracingCursorWrapper(result, self)

        logger.debug("Monkey patched SQL")



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import platform
import subprocess


def get_triple():
    return "{arch}-{platform}".format(arch=get_arch(), platform=get_platform())


def get_arch():
    """
    What CPU are we on?
    """
    arch = platform.machine()
    if arch == "i686":
        return "i686"
    elif arch == "x86_64":
        return "x86_64"
    else:
        return "unknown"


def get_platform():
    """
    What Operating System (and sub-system like glibc / musl)
    """
    system_name = platform.system()
    if system_name == "Linux":
        libc = get_libc()
        return "unknown-linux-{libc}".format(libc=libc)
    elif system_name == "Darwin":
        return "apple-darwin"
    else:
        return "unknown"


_libc = None


def get_libc():
    """
    Alpine linux uses a non glibc version of the standard library, it uses
    the stripped down musl instead. The core agent can be built against it,
    but which one is running must be detected. Shelling out to `ldd`
    appears to be the most reliable way to do this.
    """
    global _libc
    if _libc is None:
        try:
            output = subprocess.check_output(
                ["ldd", "--version"], stderr=subprocess.STDOUT, close_fds=True
            )
        except (OSError, subprocess.CalledProcessError):
            _libc = "gnu"
        else:
            if b"musl" in output:
                _libc = "musl"
            else:
                _libc = "gnu"
    return _libc



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import scout_apm.core
from scout_apm.core.config import ScoutConfig
from scout_apm.core.ignore import ignore_path
from scout_apm.core.queue_time import track_request_queue_time
from scout_apm.core.tracked_request import TrackedRequest


def includeme(config):
    configs = {}
    pyramid_config = config.get_settings()
    for name in pyramid_config:
        if name.startswith("SCOUT_"):
            value = pyramid_config[name]
            clean_name = name.replace("SCOUT_", "").lower()
            configs[clean_name] = value
    ScoutConfig.set(**configs)

    if scout_apm.core.install():
        config.add_tween("scout_apm.pyramid.instruments")


def instruments(handler, registry):
    def scout_tween(request):
        tracked_request = TrackedRequest.instance()
        span = tracked_request.start_span(operation="Controller/Pyramid")

        try:
            if ignore_path(request.path):
                tracked_request.tag("ignore_transaction", True)

            tracked_request.tag("path", request.path)

            try:
                # Determine a remote IP to associate with the request. The value is
                # spoofable by the requester so this is not suitable to use in any
                # security sensitive context.
                user_ip = (
                    request.headers.get("x-forwarded-for", default="").split(",")[0]
                    or request.headers.get("client-ip", default="").split(",")[0]
                    or request.remote_addr
                )
            except Exception:
                pass
            else:
                tracked_request.tag("user_ip", user_ip)

            try:
                queue_time = request.headers.get(
                    "x-queue-start", default=""
                ) or request.headers.get("x-request-start", default="")
            except Exception:
                pass
            else:
                track_request_queue_time(queue_time, tracked_request)

            try:
                response = handler(request)
            except Exception:
                tracked_request.tag("error", "true")
                raise

            # This happens further down the call chain. So time it starting
            # above, but only name it if it gets to here.
            if request.matched_route is not None:
                tracked_request.mark_real_request()
                span.operation = "Controller/" + request.matched_route.name

        finally:
            tracked_request.stop_span()

        return response

    return scout_tween



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

from nameko.extensions import DependencyProvider
from nameko.web.handlers import HttpRequestHandler
from werkzeug.wrappers import Request

import scout_apm.core
from scout_apm.core.ignore import ignore_path
from scout_apm.core.queue_time import track_request_queue_time
from scout_apm.core.tracked_request import TrackedRequest


class ScoutReporter(DependencyProvider):
    def setup(self):
        installed = scout_apm.core.install()
        self._do_nothing = not installed

    def worker_setup(self, worker_ctx):
        if self._do_nothing:
            return
        tracked_request = TrackedRequest.instance()
        tracked_request.mark_real_request()

        # Get HTTP details for HTTP handlers
        if isinstance(worker_ctx.entrypoint, HttpRequestHandler):
            try:
                request = worker_ctx.args[0]
            except IndexError:
                pass
            else:
                if isinstance(request, Request):
                    path = request.path
                    tracked_request.tag("path", path)

                    if ignore_path(path):
                        tracked_request.tag("ignore_transaction", True)

                    # Determine a remote IP to associate with the request. The value is
                    # spoofable by the requester so this is not suitable to use in any
                    # security sensitive context.
                    user_ip = (
                        request.headers.get("x-forwarded-for", default="").split(",")[0]
                        or request.headers.get("client-ip", default="").split(",")[0]
                        or request.remote_addr
                    )
                    tracked_request.tag("user_ip", user_ip)

                    queue_time = request.headers.get(
                        "x-queue-start", default=""
                    ) or request.headers.get("x-request-start", default="")
                    track_request_queue_time(queue_time, tracked_request)

        operation = (
            "Controller/"
            + worker_ctx.service.name
            + "."
            + worker_ctx.entrypoint.method_name
        )
        tracked_request.start_span(operation=operation)

    def worker_result(self, worker_ctx, result=None, exc_info=None):
        if self._do_nothing:
            return
        tracked_request = TrackedRequest.instance()
        if exc_info:
            tracked_request.tag("error", "true")
        tracked_request.stop_span()



# coding=utf-8
from scout_apm.core.context import AgentContext


def ignore_path(path):
    ignored_paths = AgentContext.instance.config.value("ignore")
    for ignored in ignored_paths:
        if path.startswith(ignored):
            return True
    return False



# coding=utf-8
import os
import sys
from glob import glob

from setuptools import Extension, find_packages, setup

long_description = (
    "Scout Application Performance Monitoring Agent - https://scoutapm.com"
)
if os.path.exists("README.md"):
    with open("README.md", "r") as fp:
        long_description = fp.read()

# Try to compile the extensions, except for platforms or versions
# where our extensions are not supported
compile_extensions = True

setup_args = {
    "name": "scout_apm",
    "version": "2.4.0",
    "description": "Scout Application Performance Monitoring Agent",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "url": "https://github.com/scoutapp/scout_apm_python",
    "author": "Scout",
    "author_email": "support@scoutapm.com",
    "license": "MIT",
    "zip_safe": False,
    "python_requires": ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4",
    "packages": find_packages("src"),
    "package_dir": {"": "src"},
    "py_modules": [os.splitext(os.basename(path))[0] for path in glob("src/*.py")],
    "ext_modules": [
        Extension("scout_apm.core.objtrace", ["src/scout_apm/core/ext/objtrace.c"])
    ],
    "entry_points": {
        "console_scripts": [
            "core-agent-manager = scout_apm.core.cli.core_agent_manager:main"
        ]
    },
    "install_requires": ["psutil", "requests"],
    "keywords": "apm performance monitoring development",
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "Framework :: Bottle",
        "Framework :: Django",
        "Framework :: Django :: 1.8",
        "Framework :: Django :: 1.9",
        "Framework :: Django :: 1.10",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Flask",
        "Framework :: Pyramid",
        "Intended Audience :: Developers",
        "Topic :: System :: Monitoring",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
}

if sys.version_info <= (3, 0):
    compile_extensions = False

if sys.platform.startswith("java"):
    compile_extensions = False

if "__pypy__" in sys.builtin_module_names:
    compile_extensions = False

if not compile_extensions:
    del setup_args["ext_modules"]

setup(**setup_args)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

from bottle import request

import scout_apm.core
from scout_apm.core.config import ScoutConfig
from scout_apm.core.context import AgentContext
from scout_apm.core.ignore import ignore_path
from scout_apm.core.queue_time import track_request_queue_time
from scout_apm.core.tracked_request import TrackedRequest


class ScoutPlugin(object):
    def __init__(self):
        self.name = "scout"
        self.api = 2

    def set_config_from_bottle(self, app):
        scout_config = ScoutConfig()
        bottle_configs = {}
        for k in scout_config.known_keys():
            value = app.config.get("scout.{}".format(k))
            if value is not None and value != "":
                bottle_configs[k] = value
        scout_config.set(**bottle_configs)
        return scout_config

    def setup(self, app):
        self.set_config_from_bottle(app)
        scout_apm.core.install()

    def apply(self, callback, context):
        if not AgentContext.instance.config.value("monitor"):
            return callback

        def wrapper(*args, **kwargs):
            tracked_request = TrackedRequest.instance()
            tracked_request.mark_real_request()

            if request.route.name is not None:
                path = request.route.name
            else:
                path = request.route.rule
            if path == "/":
                path = "/home"
            if not path.startswith("/"):
                path = "/{}".format(path)

            tracked_request.tag("path", path)
            if ignore_path(path):
                tracked_request.tag("ignore_transaction", True)

            tracked_request.start_span(operation="Controller{}".format(path))

            try:
                # Determine a remote IP to associate with the request. The
                # value is spoofable by the requester so this is not suitable
                # to use in any security sensitive context.
                user_ip = (
                    request.headers.get("x-forwarded-for", "").split(",")[0]
                    or request.headers.get("client-ip", "").split(",")[0]
                    or request.environ.get("REMOTE_ADDR")
                )
            except Exception:
                pass
            else:
                tracked_request.tag("user_ip", user_ip)

            try:
                queue_time = request.headers.get(
                    "x-queue-start", ""
                ) or request.headers.get("x-request-start", "")
            except Exception:
                pass
            else:
                track_request_queue_time(queue_time, tracked_request)

            try:
                return callback(*args, **kwargs)
            except Exception:
                tracked_request.tag("error", "true")
                raise
            finally:
                tracked_request.stop_span()

        return wrapper
